"""pyRserve package"""
import os
import warnings

from .rconn import connect
from .taggedContainers import TaggedList, TaggedArray, AttrArray

__version__ = open(os.path.join(os.path.dirname(__file__),
                                'version.txt')).readline().strip()

# Show all deprecated warning only once:
warnings.filterwarnings('once', category=DeprecationWarning)
del warnings



# -*- coding: utf-8 -*-
"""
Parser module for pyRserve
"""
import io
import struct
import socket

import numpy

from .rtypes import (
    CMD_OOB, CMD_RESP, DT_SEXP, DTs, ERRORS, RESP_ERR, RESP_OK,
    SOCKET_BLOCK_SIZE, VALID_R_TYPES, XT_ARRAY_BOOL, XT_ARRAY_CPLX,
    XT_ARRAY_DOUBLE, XT_ARRAY_INT, XT_ARRAY_STR, XT_BOOL, XT_CLOS, XT_DOUBLE,
    XT_HAS_ATTR, XT_INT, XT_INT3, XT_INT7, XT_LANG_NOTAG, XT_LANG_TAG, XT_LARGE,
    XT_LIST_NOTAG, XT_LIST_TAG, XT_NULL, XT_RAW, XT_S4, XT_STR, XT_SYMNAME,
    XT_UNKNOWN, XT_VECTOR, XT_VECTOR_EXP, XTs, structMap, numpyMap
)
from .misc import FunctionMapper, byteEncode, stringEncode, PY3
from .rexceptions import \
    RResponseError, REvalError, EndOfDataError, RParserError
from .taggedContainers import TaggedList, asTaggedArray, asAttrArray

DEBUG = 0


class OOBMessage(object):
    """OOB Message

    - type: OOB_SEND or OOB_MSG or OOB_STREAM_READ
    - userCode: user-defined code passed to self.oobSend/oobMessage
    - data: user-sent data or None
    - messageSize: number of bytes in user-sent data
    """
    def __init__(self, type, userCode, data=None, messageSize=0):
        self.type = type
        self.userCode = userCode
        self.data = data
        self.messageSize = messageSize

    def __len__(self):
        return self.messageSize + 16  # header


class Command(object):
    """Wrapper around the command bitfield calculating and storing its properties
    Magic extracted from RSProtocol.h
    """
    def __init__(self, code):
        self.code = code
        # Rserve 1.7 or’s the command with CMD_RESP even if it’s a OOB instead
        fixedOOBCode = code & ~CMD_RESP

        self.isOOB = bool(code & CMD_OOB)
        self.oobType = fixedOOBCode & 0x0ffff000
        self.oobUserCode = fixedOOBCode & 0xfff

        self.errCode = (code >> 24) & 127
        self.responseCode = code & 0xfffff  # lowest 20 bit


class Lexeme(list):
    """Basic Lexeme class for parsing binary data coming from Rserve"""
    def __init__(self, rTypeCode, length, hasAttr, lexpos):
        list.__init__(self, [rTypeCode, length, hasAttr, lexpos])
        self.rTypeCode = rTypeCode
        self.length = length
        self.hasAttr = hasAttr
        self.lexpos = lexpos
        self.attrLexeme = None
        self.data = None

    def setAttr(self, attrLexeme):
        self.attrLexeme = attrLexeme

    @property
    def attr(self):
        return self.attrLexeme.data if self.attrLexeme else None

    @property
    def attrLength(self):
        return self.attrLexeme.length

    @property
    def attrTypeCode(self):
        return self.attrLexeme.rTypeCode

    @property
    def dataLength(self):
        """Return length (in bytes) of actual REXPR data body"""
        if self.hasAttr:
            if not self.attrLexeme:
                raise RuntimeError('Attribute lexeme not yet set')
            # also subtract size of REXP header=4
            return self.length - self.attrLength - 4
        else:
            return self.length

    def __str__(self):
        return 'Typecode: %s   Length: %s  hasAttr: %s,  Lexpos: %d' % \
               (hex(self.rTypeCode), self.length, self.hasAttr, self.lexpos)


class Lexer(object):
    """Rserve message lexer
    Can either read a OOBMessage or a R Object
    """
    lexerMap = {}
    fmap = FunctionMapper(lexerMap)

    def __init__(self, src):
        """
        @param src: Either a string, a file object, a socket -
                    all providing valid binary r data
        """
        if type(src) == str:
            # this only works for objects implementing the buffer protocol,
            # e.g. strings, arrays, ...
            # convert string to byte object
            self.fp = io.BytesIO(byteEncode(src))
        elif type(src) == bytes:
            self.fp = io.BytesIO(src)
        else:
            self.fp = src
        if isinstance(self.fp, socket.socket):
            self._read = self.fp.recv
        else:
            self._read = self.fp.read
        # The following attributes will be set thru 'readHeader()':
        self.lexpos = None
        self.messageSize = None
        self.errCode = None
        self.responseCode = None
        self.responseOK = None
        self.isOOB = False
        self.oobType = None
        self.oobUserCode = None

    def readHeader(self):
        """
        Called initially when reading fresh data from an input source
        (file or socket). Reads header which contains data like response/error
        code and size of data entire package.

        QAP1 header structure parts (16 bytes total):

            [ 0-3 ] (int) command
            [ 4-7 ] (int) length of the message (bits 0-31)
            [ 8-11] (int) offset of the data part
            [12-15] (int) length of the message (bits 32-63)
        """
        self.lexpos = 0

        command = Command(struct.unpack('<I', self.read(4))[0])
        # Obtain lower 32bit part of message length:
        messageSize1 = self.__unpack(XT_INT)
        dataOffset = self.__unpack(XT_INT)
        assert dataOffset == 0, 'dataOffset > 0 is not implemented'
        # Obtain upper 32bit part of message length:
        messageSize2 = self.__unpack(XT_INT) << 32  # shift 32bits to the left
        self.messageSize = messageSize2 + messageSize1

        self.isOOB = command.isOOB
        if self.isOOB:
            # FIXME: Rserve has a bug(?) that sets CMD_RESP on
            #        OOB commands so we clear it for now
            self.oobType = command.oobType
            self.oobUserCode = command.oobUserCode

            if DEBUG:
                print('oob type: %x, oob user code: %x, message size: %d' %
                      (self.oobType, self.oobUserCode, self.messageSize))
        else:
            self.errCode = command.errCode

            self.responseCode = command.responseCode
            if self.responseCode == RESP_OK:
                self.responseOK = True
            elif self.responseCode == RESP_ERR:
                self.responseOK = False
            else:
                self.clearSocketData()
                raise ValueError('Received illegal response code (%x)' %
                                 self.responseCode)

            if DEBUG:
                print('response ok? %s (responseCode=%x), error-code: %x, '
                      'message size: %d' %
                      (self.responseOK, self.responseCode,
                       self.errCode, self.messageSize))

        return self.messageSize

    def clearSocketData(self):
        """
        If for any reason the parsing process returns an error, make sure that
        all data from a socket is removed to avoid data pollution with further
        parsing attempts.
        """
        if not isinstance(self.fp, socket.socket):
            # not a socket. Nothing to do here.
            return
        # Switch socket into non-blocking mode and read from it until it
        # is empty (and hence socket.error is raised):
        self.fp.setblocking(False)
        try:
            while True:
                self.fp.recv(SOCKET_BLOCK_SIZE)
        except socket.error:
            # socket has no more data, it can be considered as cleared
            pass
        finally:
            # Now set it back to blocking mode (no matter what exception):
            self.fp.setblocking(True)

    def read(self, length):
        """
        Read number of bytes from input data source (file or socket).
        If end of data is reached it raises EndOfDataError().

        Sockets might not return all requested data at once, so use an io
        buffer to collect all data needed in a loop.
        """
        bytesToRead = length
        buf = io.BytesIO(b'')
        while bytesToRead > 0:
            fragment = self._read(bytesToRead)
            lenFrag = len(fragment)
            if lenFrag == 0:
                raise EndOfDataError()
            buf.write(fragment)
            bytesToRead -= lenFrag

        self.lexpos += length
        data = buf.getvalue()
        return data

    def __unpack(self, tCode, num=None):
        """
        Read 'num' (atomic) data items from the input source and convert them
        into a list of python objects. Byteswapping for numeric data will
        be done.
        """
        structCode = structMap[tCode] if type(tCode) == int else tCode
        # All data from Rserve is stored in little-endian format!
        fmt = byteEncode('<' + str(num) + structCode if (num is not None)
                         else '<' + structCode)  # convert into bytes!
        if tCode == XT_INT3:
            length = 3
            rawData = self.read(length) + b'\x00'
        elif tCode == XT_INT7:
            length = 7
            rawData = self.read(length) + b'\x00'
        else:
            length = struct.calcsize(fmt or 1)
            rawData = self.read(length)
        d = struct.unpack(fmt, rawData)
        return d[0] if num is None else list(d)

    def nextExprHdr(self):
        """
        From the input file/socket determine the type of the next data item,
        and its length.
        This method can be applied to read the
        - entire data header (containing one of the DT_* codes)
        - an REXPR header
        """
        startLexpos = self.lexpos
        _rTypeCode = self.__unpack('B')  # unsigned byte!
        # extract pure rTypeCode without XT_HAS_ATTR or XT_LARGE flags:
        rTypeCode = _rTypeCode & 0x3F
        # extract XT_HAS_ATTR flag (if it exists)"
        hasAttr = (_rTypeCode & XT_HAS_ATTR) != 0
        # extract XT_LARGE flag (if it exists):
        isXtLarge = (_rTypeCode & XT_LARGE) != 0
        if isXtLarge:
            # header is larger, use all 7 bytes for length information
            # (new in Rserve 0.3)
            length = self.__unpack(XT_INT7)
        else:
            # small header, use 3 bytes for length information
            length = self.__unpack(XT_INT3)
        if rTypeCode not in VALID_R_TYPES:
            raise RParserError(
                "Unknown SEXP type %s found at lexpos %d, length %d" %
                (hex(rTypeCode), startLexpos, length))
        return Lexeme(rTypeCode, length, hasAttr, startLexpos)

    def nextExprData(self, lexeme):
        """
        Read next data item from binary r data and transform it into a
        python object.
        """
        return self.lexerMap[lexeme.rTypeCode](self, lexeme)

    ###########################################################################

    @fmap(XT_INT, XT_DOUBLE)
    def xt_atom(self, lexeme):
        raw = self.read(lexeme.dataLength)
        return struct.unpack(
            byteEncode('<%s' % structMap[lexeme.rTypeCode]), raw)[0]

    @fmap(XT_BOOL)
    def xt_bool(self, lexeme):
        raw = self.read(lexeme.dataLength)
        # a boolean is stored in a 4 bytes word, but only the first byte
        # is significant:
        if PY3:
            # python3 directly converts a single byte item into a number!
            b = raw[0]
        else:
            b = struct.unpack(byteEncode('<%s' % structMap[XT_BOOL]),
                              raw[0])[0]
            # b can be 2, meaning NA. Otherwise transform 0/1 into False/True
        return None if b == 2 else b == 1

    @fmap(XT_ARRAY_INT, XT_ARRAY_DOUBLE, XT_ARRAY_CPLX)
    def xt_array_numeric(self, lexeme):
        raw = self.read(lexeme.dataLength)
        # TODO: swapping...
        data = numpy.frombuffer(raw, dtype=numpyMap[lexeme.rTypeCode])
        return data

    @fmap(XT_ARRAY_BOOL)
    def xt_array_bool(self, lexeme):
        """A boolean array consists of a 4-byte word (i.e. integer)
        determining the number of boolean values in the following dataLength-4
        bytes.
        E.g. a bool array of one TRUE item looks like:
        01 00 00 00   01 ff ff ff

        The first 01 value tells that there is one bool value in the array.
        The other 01 is the TRUE value, the other 3 'ff's are padding bytes.
        Those will be used if the vector has 2,3 or 4 boolean values.
        For a fifth value another 4 bytes are appended.
        """
        numBools = self.__unpack(XT_INT, 1)[0]
        # read the actual boolean values, including padding bytes:
        raw = self.read(lexeme.dataLength - 4)
        # Check if the array contains any NA values (encoded as \x02).
        # If so we need to convert the 2's to None's and use a numpy
        # array of type Object otherwise numpy will cast the None's into False's.
        # This is handled for us for numeric types since numpy can use it's own
        # nan type, but here we need to help it out.
        if 2 in raw:
            data = numpy.frombuffer(raw[:numBools], dtype=numpy.int8).astype(object)
            data[data == 2] = None
        else:
            data = numpy.frombuffer(
                raw[:numBools],
                dtype=numpyMap[lexeme.rTypeCode]
            )
        return data

    @fmap(XT_ARRAY_STR)
    def xt_array_str(self, lexeme):
        """
        An array of one or more null-terminated strings.
        The XT_ARRAY_STR can contain trailing chars \x01 which need to be
        chopped off. Since strings are encoded as bytes (in Py3) they need
        to be converted into real strings.
        """
        if lexeme.dataLength == 0:
            return ''
        raw = self.read(lexeme.dataLength)
        bytesStrList = raw.split(b'\0')[:-1]
        strList = [stringEncode(byteString) for byteString in bytesStrList]
        return numpy.array(strList)

    @fmap(XT_STR)
    def xt_str(self, lexeme):
        """
        A null-terminated string.
        It's length can be larger than the actual string since it is always a
        multiple of 4.
        The rest is filled with trailing \0s which need to be chopped off.
        """
        raw = self.read(lexeme.dataLength)
        byteStr = raw.split(b'\0', 1)[0]
        return stringEncode(byteStr)

    @fmap(XT_SYMNAME)
    def xt_symname(self, lexeme):
        """
        Just like a string, but in S4 classes, a special value for NULL exists
        """
        string = self.xt_str(lexeme)
        return None if string == '\x01NULL\x01' else string

    @fmap(XT_NULL)
    def xt_null(self, lexeme):
        return None

    @fmap(XT_UNKNOWN)
    def xt_unknown(self, lexeme):
        return self.__unpack(XT_INT)

    @fmap(XT_RAW)
    def xt_raw(self, lexeme):
        self.__unpack(XT_INT)
        return self.read(lexeme.dataLength - 4)


class RParser(object):
    #
    parserMap = {}
    fmap = FunctionMapper(parserMap)

    def __init__(self, src, atomicArray):
        """
        atomicArray: if False parsing arrays with only one element will just
                     return this element
        arrayOrder:  The order in which data in multi-dimensional arrays is
                     returned. 'C' for c-order, F for fortran.
        """
        self.lexer = Lexer(src)
        self.atomicArray = atomicArray
        self.indentLevel = None

    def __getitem__(self, key):
        return self.parserMap[key]

    def __getattr__(self, attr):
        if attr in ['messageSize']:
            return getattr(self.lexer, attr)
        else:
            raise AttributeError(attr)

    @property
    def __ind(self):
        # return string with number of spaces appropriate for current
        # indentation level
        return self.indentLevel * 4 * ' '

    def _debugLog(self, lexeme, isRexpr=True):
        if DEBUG:
            lx = lexeme
            typeCodeDict = XTs if isRexpr else DTs
            print('%s %s (%s), hasAttr=%s, lexpos=%d, length=%s' %
                  (self.__ind, typeCodeDict[lx.rTypeCode], hex(lx.rTypeCode),
                   lx.hasAttr, lx.lexpos, lx.length))

    def parse(self):
        """
        Parse data stream and return result converted into
        python data structure
        """
        self.indentLevel = 1
        self.lexer.readHeader()

        message = None
        if self.lexer.messageSize > 0:
            try:
                message = self._parse()
            except Exception:
                # If any error is raised during lexing and parsing, make sure
                # that the entire data is read from the input source if it is
                # a socket, otherwise following attempts to
                # parse again from a socket will return polluted data:
                self.lexer.clearSocketData()
                raise
        elif not self.lexer.responseOK:
            try:
                rserve_err_msg = ERRORS[self.lexer.errCode]
            except KeyError:
                raise REvalError("R evaluation error (code=%d)" %
                                 self.lexer.errCode)
            else:
                raise RResponseError('Response error %s (error code=%d)' %
                                     (rserve_err_msg, self.lexer.errCode))

        if self.lexer.isOOB:
            return OOBMessage(self.lexer.oobType, self.lexer.oobUserCode,
                              message, self.lexer.messageSize)
        else:
            return message

    def _parse(self):
        dataLexeme = self.lexer.nextExprHdr()
        self._debugLog(dataLexeme, isRexpr=False)
        if dataLexeme.rTypeCode == DT_SEXP:
            lexeme = self._parseExpr()
            return self._postprocessData(lexeme.data)
        else:
            raise NotImplementedError()

    def _parseExpr(self):
        self.indentLevel += 1
        lexeme = self.lexer.nextExprHdr()
        self._debugLog(lexeme)
        if lexeme.hasAttr:
            self.indentLevel += 1
            if DEBUG:
                print('%s Attribute:' % self.__ind)
            lexeme.setAttr(self._parseExpr())
            self.indentLevel -= 1
        lexeme.data = self.parserMap.get(lexeme.rTypeCode,
                                         self[None])(self, lexeme)
        self.indentLevel -= 1
        return lexeme

    def _nextExprData(self, lexeme):
        lexpos = self.lexer.lexpos
        data = self.lexer.nextExprData(lexeme)
        if DEBUG:
            print('%s    data-lexpos: %d, data-length: %d bytes' %
                  (self.__ind, lexpos, lexeme.dataLength))
            print('%s    data: %s' % (self.__ind, repr(data)))
            try:
                dataLen = len(data)
                print('%s    length: %d' % (self.__ind, dataLen))
            except TypeError:
                pass
        return data

    def _postprocessData(self, data):
        """
        Postprocess parsing results depending on configuration parameters
        Currently only arrays are effected.
        """
        if data.__class__ == numpy.ndarray:
            # this does not apply for arrays with attributes
            # (__class__ would be TaggedArray)!
            if len(data) == 1 and not self.atomicArray:
                # if data is a plain numpy array, and has only one element,
                # just extract and return this.
                # For convenience reasons type-convert it into a native
                # Python data type:
                data = data[0]
                if isinstance(data, (float, numpy.float64)):
                    # convert into native python float:
                    data = float(data)
                elif isinstance(data, (int, numpy.int32, numpy.int64)):
                    # convert into native int or long, depending on value:
                    data = int(data)
                elif isinstance(data, (complex, numpy.complex64,
                                       numpy.complex128)):
                    # convert into native python complex number:
                    data = complex(data)
                elif isinstance(data, (numpy.string_, str)):
                    # convert into native python string:
                    data = str(data)
                elif isinstance(data, (bool, numpy.bool8)):
                    # convert into native python string
                    data = bool(data)
        return data

    @fmap(None)
    def xt_(self, lexeme):
        # apply this for atomic data
        return self._nextExprData(lexeme)

    @fmap(XT_ARRAY_BOOL, XT_ARRAY_INT, XT_ARRAY_DOUBLE, XT_ARRAY_STR)
    def xt_array(self, lexeme):
        # converts data into a numpy array already:
        data = self._nextExprData(lexeme)
        if lexeme.hasAttr and lexeme.attrTypeCode == XT_LIST_TAG:
            for tag, value in lexeme.attr:
                if tag == 'dim':
                    # the array has a defined shape, and R stores and
                    # sends arrays in Fortran mode:
                    data = data.reshape(value, order='F')
                elif tag == 'names':
                    # convert numpy-vector 'value' into list to make
                    # TaggedArray work properly:
                    data = asTaggedArray(data, list(value))
                else:
                    # there are additional tags in the attribute, just collect
                    # them in a dictionary attached to the array.
                    try:
                        data.attr[tag] = value
                    except AttributeError:
                        data = asAttrArray(data, {tag: value})
        return data

    @fmap(XT_VECTOR, XT_VECTOR_EXP, XT_LANG_NOTAG, XT_LIST_NOTAG)
    def xt_vector(self, lexeme):
        """
        A vector is e.g. return when sending "list('abc','def')" to R. It can
        contain mixed types of data items.
        The binary representation of an XT_VECTOR is weird: a vector contains
        unknown number of items, with possibly variable length. Only the number
        of bytes of the data of a vector is known in advance.
        The end of this REXP can only be detected by keeping track of how
        many bytes have been consumed (lexeme.length!) until the end of the
        REXP has been reached.

        A vector expression (type 0x1a) is according to Rserve docs the same
        as XT_VECTOR. For now just a list with the expression content is
        returned in this case.
        """
        finalLexpos = self.lexer.lexpos + lexeme.dataLength
        if DEBUG:
            print('%s     Vector-lexpos: %d, length %d, finished at: %d' %
                  (self.__ind, self.lexer.lexpos,
                   lexeme.dataLength, finalLexpos))
        data = []
        while self.lexer.lexpos < finalLexpos:
            # convert single item arrays into atoms (via stripArray)
            data.append(self._postprocessData(self._parseExpr().data))

        if lexeme.hasAttr and lexeme.attrTypeCode == XT_LIST_TAG:
            # The vector is actually a tagged list, i.e. a list which allows
            # to access its items by name (like in a dictionary). However items
            # are ordered, and there is not necessarily a name available for
            # every item.
            for tag, value in lexeme.attr:
                if tag == 'names':
                    # the vector has named items
                    data = TaggedList(zip(value, data))
                else:
                    if DEBUG:
                        print('Warning: applying LIST_TAG "%s" on xt_vector '
                              'not yet implemented' % tag)
        return data

    @fmap(XT_LIST_TAG, XT_LANG_TAG)
    def xt_list_tag(self, lexeme):
        # a xt_list_tag usually occurs as an attribute of a vector or list
        # (like for a tagged list)
        finalLexpos = self.lexer.lexpos + lexeme.dataLength
        r = []
        while self.lexer.lexpos < finalLexpos:
            value, tag = self._parseExpr().data, self._parseExpr().data
            # reverse order of tag and value when adding it to result list
            r.append((tag, value))
        return r

    @fmap(XT_CLOS)
    def xt_closure(self, lexeme):
        # read entire data provided for closure (a R code object) even though
        # we don't know what to do with it on the Python side ;-)
        aList1 = self._parseExpr().data
        aList2 = self._parseExpr().data
        # Some closures seem to provide their sourcecode in an attrLexeme,
        # but some don't.
        # return Closure(lexeme.attrLexeme.data[0][1])
        # So for now let's just return the entire parse tree in a
        # Closure instance.
        return Closure(lexeme, aList1, aList2)

    @fmap(XT_S4)
    def xt_s4(self, lexeme):
        """A S4 object only contains attributes, no other payload"""
        if lexeme.hasAttr and lexeme.attrTypeCode == XT_LIST_TAG:
            return S4(lexeme.attr)
        else:
            return S4([])


##############################################################################


def rparse(src, atomicArray=False):
    rparser = RParser(src, atomicArray)
    return rparser.parse()

##############################################################################


class Closure(object):
    """
    Very simple container to return "something" for a closure.
    Not really usable in Python though.
    """
    def __init__(self, lexeme, aList1, aList2):
        self.lexeme = lexeme
        self.aList1 = aList1
        self.aList2 = aList2

    def __repr__(self):
        return '<Closure instance %d>' % id(self)


class S4(dict):
    """Very simple representation of a S4 instance"""
    def __init__(self, source=(), **entries):
        super(S4, self).__init__(source, **entries)

        if 'class' in self:
            self.classes = self['class']
            del self['class']
        else:
            self.classes = []

    def __repr__(self):
        attrs = super(S4, self).__repr__()
        return "<S4 classes={} {}>".format(self.classes, attrs)



"""
Serializer class to convert Python objects into a binary data stream for
sending them to Rserve.
"""
import os
import io
import struct
import socket

import numpy

from . import rtypes
from .misc import PY3, FunctionMapper, byteEncode, padLen4, string2bytesPad4
from .taggedContainers import TaggedList, TaggedArray

# turn on DEBUG to see extra information about what the serializer is
# doing with your data
DEBUG = 0

NoneType = type(None)

if PY3:
    # make test work with Python 3 where 'long'-type does not exist:
    long = int


class RSerializer(object):
    """
    Class to to serialize Python objects into a binary data stream for sending
    them to Rserve.

    Depending on 'commandType' given to __init__ the resulting binary string
    can be used to send a command, to assign a variable in Rserve, or to
    reply to a request received from Rserve.
    """
    serializeMap = {}
    fmap = FunctionMapper(serializeMap)

    def __init__(self, commandType, fp=None):
        if isinstance(fp, socket.socket):
            # kwargs = {'mode': 'b'} if PY3 else {}
            self._fp = fp
            self._buffer = io.BytesIO()
        elif not fp:
            self._buffer = fp or io.BytesIO()
            self._fp = None
        else:
            # expect fp to be a seekable file(-like) object:
            self._buffer = self._fp = fp
        self._dataSize = 0
        self._writeHeader(commandType)

    def _getRetVal(self):
        if self._fp is self._buffer:
            # file(-like) object - data has been written, nothing to return
            return None
        elif not self._fp:
            # data has only been written into buffer, so return its value:
            return self._buffer.getvalue()
        else:
            # i.e. socket: write result of _fp into socket-fp
            self._fp.send(self._buffer.getvalue())
            return None

    def _writeHeader(self, commandType):
        """Write main header of message for Rserve"""
        # Set length to zero initially, will be fixed in _finalizerHeader()
        # when msg size is determined:
        msg_length_lower = msg_length_higher = 0
        data_offset = 0
        header = struct.pack('<IIII', commandType, msg_length_lower,
                             data_offset, msg_length_higher)
        if DEBUG:
            print('Writing header: %d bytes: %s' % (len(header), repr(header)))
        self._buffer.write(header)

    def finalize(self):
        """Finalize the message package before actually sending/wriring it out.
        -> Set the length of the entire data package in the general message hdr
           as number of bytes of the entire message minus the general hdr
        """
        # Jump to end of buffer to determine its length:
        self._buffer.seek(0, os.SEEK_END)
        messageSize = self._buffer.tell() - rtypes.RHEADER_SIZE
        if DEBUG:
            print('writing size of header: %2d' % messageSize)
        # Goto position 4 of the general Rserve package header and write the
        # size of the overall rserve message there. For message size > 2**32
        # the size is split into two parts, the lower 32 bits are written at
        # position 4, the higher part is written at position 12 (see QAP1 docs)
        bin_messageSize = struct.pack('<Q', messageSize)
        bin_messageSize_lo = bin_messageSize[:4]
        bin_messageSize_hi = bin_messageSize[4:]

        self._buffer.seek(4)
        self._buffer.write(bin_messageSize_lo)
        self._buffer.write(b'\x00\x00\x00\x00')  # data offset, zero by default
        self._buffer.write(bin_messageSize_hi)
        return self._getRetVal()

    def _writeDataHeader(self, rTypeCode, length):
        """Write a header for either DataTypes (DT_*) or ExpressionTypes (XT_*)

        According to the documentation of Rserve:
        -----------------------------------------
        If the length of the data block is smaller than 2**24 - 16 (fffff0)
        then the header has a length of 4 bytes and looks like:
            [1]   rTypeCode
            [2-4] length of data block (3 bytes/24 bits)
        If the length of the data is larger, then the rTypeCode of the header
        has to be OR'ed with XT_LARGE (or DT_LARGE, which is the same). Then
        the length of the datablock can be encoded in 7 bytes:
            [1]   rTypeCode
            [2-4] length of data block (lower three bytes)
            [5-8] length of data block (upper four bytes)

        However; pyRserve is not capable of dynamic header sizes, so we will
                 use the large header setup for all data packages no matter of
                 their size. Simon Urbanek confirmed that this does not cause
                 any problems with Rserve.
        """
        rTypeCode |= rtypes.XT_LARGE
        hdr = struct.pack('<BQ', rTypeCode, length)
        # cut-off leftover zeros at the right end of the header string
        # before writing it to the buffer:
        self._buffer.write(hdr[:rtypes.LARGE_DATA_HEADER_SIZE])
        return rtypes.LARGE_DATA_HEADER_SIZE

    def serialize(self, o, dtTypeCode=rtypes.DT_SEXP):
        # Here the data typecode (DT_* ) of the entire message is written,
        # with its length. Then the actual data itself is written out.
        if dtTypeCode == rtypes.DT_STRING:
            paddedString = string2bytesPad4(o)
            length = len(paddedString)
            hdrSize = self._writeDataHeader(dtTypeCode, length)
            self._buffer.write(paddedString)
        elif dtTypeCode == rtypes.DT_INT:
            length = 4   # an integer is encoded as 4 bytes
            hdrSize = self._writeDataHeader(dtTypeCode, length)
            self._buffer.write(struct.pack('<i', o))
        elif dtTypeCode == rtypes.DT_SEXP:
            startPos = self._buffer.tell()
            self._buffer.write(b'\0\0\0\0\0\0\0\0')
            length = self.serializeExpr(o)
            self._buffer.seek(startPos)
            hdrSize = self._writeDataHeader(dtTypeCode, length)
        else:
            raise NotImplementedError('no support for DT-type %x' % dtTypeCode)
        # Jump back to end of buffer to be prepared for writing more data
        self._buffer.seek(0, os.SEEK_END)
        # Adjust datasize counter
        self._dataSize += length + hdrSize

    def serializeExpr(self, o):
        if isinstance(o, numpy.ndarray):
            rTypeCode = rtypes.numpyMap[o.dtype.type]
        else:
            rTypeCode = type(o)
        try:
            s_func = self.serializeMap[rTypeCode]
        except KeyError:
            raise NotImplementedError('Serialization of "%s" not implemented' %
                                      rTypeCode)
        startPos = self._buffer.tell()
        if DEBUG:
            print('Serializing expr %r with rTypeCode=%s using function %s' %
                  (o, rTypeCode, s_func))
        s_func(self, o)
        # determine and return the length of actual R expression data:
        return self._buffer.tell() - startPos

    @fmap(NoneType, rtypes.XT_NULL)
    def s_null(self, _):
        """Send Python's None to R, resulting in NULL there"""
        # For NULL only the header needs to be written, there is no data body.
        self._writeDataHeader(rtypes.XT_NULL, 4)

    @fmap(rtypes.XT_STR, rtypes.XT_SYMNAME)
    def s_string_or_symbol(self, o, rTypeCode=rtypes.XT_STR):
        """
        Possible rTypeCodes for a given string are:
        - XT_STR
        - XT_SYMNAME
        """
        # The string packet contains trailing padding zeros to make it always
        # a multiple of 4 in length:
        paddedString = string2bytesPad4(o)
        length = len(paddedString)
        self._writeDataHeader(rTypeCode, length)
        if DEBUG:
            print('Writing string: %2d bytes: %s' %
                  (length, repr(paddedString)))
        self._buffer.write(paddedString)

    # ############### Arrays #########################################

    def __s_write_xt_array_tag_data(self, o):
        """
        Write tag data of an array, like dimension for a multi-dim array,
        or other information found. Return appropriate rTypeCode.
        """
        xt_tag_list = []
        if o.ndim > 1:
            xt_tag_list.append((b'dim', numpy.array(o.shape, numpy.int32)))
        if isinstance(o, TaggedArray):
            xt_tag_list.append((b'names', numpy.array(o.attr)))

        attrFlag = rtypes.XT_HAS_ATTR if xt_tag_list else 0
        rTypeCode = rtypes.numpyMap[o.dtype.type] | attrFlag
        # write length of zero for now, will be corrected later:
        self._writeDataHeader(rTypeCode, 0)
        if attrFlag:
            self.s_xt_tag_list(xt_tag_list)
        return rTypeCode

    def __s_update_xt_array_header(self, headerPos, rTypeCode):
        """
        Update length information of xt array header which has been
        previously temporarily set to 0 in __s_write_xt_array_tag_data()
        @arg headerPos: file position where header information should be
                        written.
        @arg rTypeCode
        """
        # subtract length of data header (8 bytes), does not count to payload!
        length = self._buffer.tell() - headerPos - rtypes.LARGE_DATA_HEADER_SIZE
        self._buffer.seek(headerPos)
        self._writeDataHeader(rTypeCode, length)
        self._buffer.seek(0, os.SEEK_END)

    @fmap(*rtypes.STRING_TYPES)
    def s_xt_array_single_str(self, o):
        """Serialize single string object"""
        arr = numpy.array([o])
        self.s_xt_array_str(arr)

    @fmap(rtypes.XT_ARRAY_STR)
    def s_xt_array_str(self, o):
        """Serialize array of strings"""
        startPos = self._buffer.tell()
        rTypeCode = self.__s_write_xt_array_tag_data(o)

        # reshape into 1d array:
        o1d = o.reshape(o.size, order='F')
        # Byte-encode them:
        bo = [byteEncode(d) for d in o1d]
        # add empty string to that the following join with \0 adds an
        # additional zero at the end of the last string!
        bo.append(b'')
        # Concatenate them as null-terminated strings:
        nullTerminatedStrings = b'\0'.join(bo)

        padLength = padLen4(nullTerminatedStrings)
        self._buffer.write(nullTerminatedStrings)
        self._buffer.write(b'\1\1\1\1'[:padLength])

        # Update the array header:
        self.__s_update_xt_array_header(startPos, rTypeCode)

    @fmap(bool, numpy.bool_)
    def s_atom_to_xt_array_boolean(self, o):
        """
        Render single boolean items into their corresponding array
        counterpart in R.
        Always convert a boolean atomic value into a specialized boolean
        R vector.
        """
        arr = numpy.array([o])
        self.s_xt_array_boolean(arr)

    @fmap(rtypes.XT_ARRAY_BOOL)
    def s_xt_array_boolean(self, o):
        """
        - o: numpy array or subclass (e.g. TaggedArray) with boolean values
        Note: If o is multi-dimensional a tagged array is created. Also if o
              is of type TaggedArray.
        """
        startPos = self._buffer.tell()
        rTypeCode = self.__s_write_xt_array_tag_data(o)

        # A boolean vector starts with its number of boolean values in the
        # vector (as int32):
        structCode = '<'+rtypes.structMap[int]
        self._buffer.write(struct.pack(structCode, o.size))
        # Then write the boolean values themselves. Note that R expects binary
        # array data in Fortran order, so prepare this accordingly:
        data = o.tobytes(order='F')
        self._buffer.write(data)
        # Finally pad the binary data to be of a multiple of four in length:
        self._buffer.write(padLen4(data) * b'\xff')

        # Update the array header:
        self.__s_update_xt_array_header(startPos, rTypeCode)

    @fmap(int, numpy.int32, long, numpy.int64, numpy.compat.long, float, complex,
          numpy.float64, numpy.complex64, numpy.complex128)
    def s_atom_to_xt_array_numeric(self, o):
        """
        Render single numeric items into their corresponding array counterpart
        in R
        """
        if isinstance(o, (int, long, numpy.int64, numpy.compat.long)):
            if rtypes.MIN_INT32 <= o <= rtypes.MAX_INT32:
                # even though this type of data is 'long' it still fits into a
                # normal integer. Good!
                o = int(o)
            else:
                raise ValueError('Cannot serialize long integers larger than '
                                 'MAX_INT32 (**31-1)')

        rTypeCode = rtypes.atom2ArrMap[type(o)]
        structCode = '<'+rtypes.structMap[type(o)]
        length = struct.calcsize(structCode)
        if type(o) is complex:
            self._writeDataHeader(rTypeCode, length*2)
            self._buffer.write(struct.pack(structCode, o.real))
            self._buffer.write(struct.pack(structCode, o.imag))
        else:
            self._writeDataHeader(rTypeCode, length)
            self._buffer.write(struct.pack(structCode, o))

    @fmap(rtypes.XT_ARRAY_CPLX, rtypes.XT_ARRAY_DOUBLE, rtypes.XT_ARRAY_INT)
    def s_xt_array_numeric(self, o):
        """
        @param o: numpy array or subclass (e.g. TaggedArray)
        @note: If o is multi-dimensional a tagged array is created. Also if o
               is of type TaggedArray.
        """
        if o.dtype in (numpy.int64, numpy.compat.long):
            # Note: use int instead of compat.long once Py2 is abandoned.
            if rtypes.MIN_INT32 <= o.min() and o.max() <= rtypes.MAX_INT32:
                # even though this type of array is 'long' its values still
                # fit into a normal int32 array. Good!
                o = o.astype(numpy.int32)
            else:
                raise ValueError('Cannot serialize long integer arrays with '
                                 'values outside MAX_INT32 (2**31-1) range')

        startPos = self._buffer.tell()
        rTypeCode = self.__s_write_xt_array_tag_data(o)

        # TODO: make this also work on big endian machines (data must be
        #       written in little-endian!!)

        # Note: R expects binary array data in Fortran order, so prepare this
        # accordingly:
        self._buffer.write(o.tobytes(order='F'))

        # Update the array header:
        self.__s_update_xt_array_header(startPos, rTypeCode)

    # ############## Vectors and Tag lists ####################################

    @fmap(list, TaggedList)
    def s_xt_vector(self, o):
        """Render all objects of given python list into generic r vector"""
        startPos = self._buffer.tell()
        # remember start position for calculating length in bytes of entire
        # list content
        attrFlag = rtypes.XT_HAS_ATTR if o.__class__ == TaggedList else 0
        self._writeDataHeader(rtypes.XT_VECTOR | attrFlag, 0)
        if attrFlag:
            self.s_xt_tag_list([(b'names', numpy.array(o.keys))])
        for v in o:
            self.serializeExpr(v)
        length = self._buffer.tell() - startPos
        self._buffer.seek(startPos)
        # now write header again with correct length information
        # subtract length of list data header:
        self._writeDataHeader(rtypes.XT_VECTOR | attrFlag,
                              length - rtypes.LARGE_DATA_HEADER_SIZE)
        self._buffer.seek(0, os.SEEK_END)

    def s_xt_tag_list(self, o):
        startPos = self._buffer.tell()
        self._writeDataHeader(rtypes.XT_LIST_TAG, 0)
        for tag, data in o:
            self.serializeExpr(data)
            self.s_string_or_symbol(tag, rTypeCode=rtypes.XT_SYMNAME)
        length = self._buffer.tell() - startPos
        self._buffer.seek(startPos)
        # now write header again with correct length information
        # subtract length of list data header:
        self._writeDataHeader(rtypes.XT_LIST_TAG,
                              length - rtypes.LARGE_DATA_HEADER_SIZE)
        self._buffer.seek(0, os.SEEK_END)

    # ##########################################################
    # ### class methods for calling specific Rserv functions ###

    @classmethod
    def rEval(cls, aString, fp=None, void=False):
        """
        Create binary code for evaluating a string expression remotely in
        Rserve
        """
        cmd = rtypes.CMD_voidEval if void else rtypes.CMD_eval
        s = cls(cmd, fp=fp)
        s.serialize(aString, dtTypeCode=rtypes.DT_STRING)
        return s.finalize()

    @classmethod
    def rAssign(cls, varname, o, fp=None):
        """
        Create binary code for assigning an expression to a variable remotely
        in Rserve
        """
        s = cls(rtypes.CMD_setSEXP, fp=fp)
        s.serialize(varname, dtTypeCode=rtypes.DT_STRING)
        s.serialize(o, dtTypeCode=rtypes.DT_SEXP)
        return s.finalize()

    @classmethod
    def rShutdown(cls, fp=None):
        s = cls(rtypes.CMD_shutdown, fp=fp)
        return s.finalize()

    @classmethod
    def rSerializeResponse(cls, Rexp, fp=None):
        # mainly used for unittesting
        s = cls(rtypes.RESP_OK, fp=fp)
        s.serialize(Rexp, dtTypeCode=rtypes.DT_SEXP)
        return s.finalize()


# Some shortcuts:
rEval = RSerializer.rEval
rAssign = RSerializer.rAssign
rSerializeResponse = RSerializer.rSerializeResponse
rShutdown = RSerializer.rShutdown



import os
from setuptools import setup
from pyRserve import __version__

PACKAGE_NAME = "pyRserve"

requirements = open('requirements.txt').read().splitlines()
requirements_testing = open('requirements_dev.txt').read().splitlines()

# Get long_description from intro.txt:
here = os.path.dirname(os.path.abspath(__file__))
with open('README.rst') as fp:
    long_description = fp.read()

setup(
    name=PACKAGE_NAME,
    version=__version__,
    description='A Python client to remotely access the R statistic package '
                'via network',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author='Ralph Heinkel',
    author_email='rh@ralph-heinkel.com',
    url='https://pypi.org/project/pyRserve/',
    project_urls={
        'Documentation': 'https://pyrserve.readthedocs.io/',
        'Changelog': 'https://pyrserve.readthedocs.io/en/latest/changelog.html',
        'Source': 'https://github.com/ralhei/pyRserve',
        'Tracker': 'https://github.com/ralhei/pyRserve/issues',
    },
    keywords='R Rserve',
    packages=['pyRserve'],
    install_requires=requirements,
    extras_require={
        'testing': requirements_testing
    },
    include_package_data=True,
    license='MIT license',
    platforms=['unix', 'linux', 'cygwin', 'win32'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Networking',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
)
